from spdx.checksum import Checksum,ChecksumAlgorithm
from spdx.utils import NoAssert
from spdx.creationinfo import Person
from spdx.license import License
from spdx.document import (Document,ExternalDocumentRef)
from spdx.package import Package
from spdx.version import Version

master_doc = Document()

class SPDX_ShallowMerger():
    def __init__(self,doc_list=None,docnamespace=None,name=None,author=None,email=None):
        self.doc_list = doc_list
        self.docnamespace = docnamespace
        self.name = name
        self.author = author
        self.emailaddr = email

    def get_document(self):
        return master_doc

    def doc_creationInfo(self):
        master_doc.name = self.name
        master_doc.version = Version(2,3) # TODO Need to check from where to take this. can not hardcode here
        master_doc.spdx_id = self.docnamespace + "#SPDXRef-DOCUMENT"
        master_doc.namespace = self.docnamespace
        master_doc.data_license = License.from_identifier("CC0-1.0") #TODO Can not hardcode it here need to check from where to take it.
        master_doc.creation_info.add_creator(Person(self.author,self.emailaddr))
        master_doc.creation_info.set_created_now()

    def doc_externalDocumentRef(self):
        package = Package()
        package.name = self.name
        package.version = "1.0"
        package.spdx_id = self.docnamespace + "#SPDXRef-DOCUMENT"
        package.download_location = NoAssert()

        master_doc.add_package(package)
        for doc in self.doc_list:
            check_sum = Checksum(ChecksumAlgorithm.SHA1,doc.comment)
            extDoc = ExternalDocumentRef(doc.spdx_id,doc.namespace,check_sum)
            master_doc.add_ext_document_reference(extDoc)



import os
from spdx.parsers import parse_anything
from spdxmerge.checksum import sha1sum

def read_docs(dir):
    doc_list = []
    doc_files = [f for f in os.listdir(dir) if f.endswith(('.json', '.spdx'))]
    for file in doc_files:
        doc, _error = parse_anything.parse_file(dir+"/"+file)
        check_sum = sha1sum(dir+"/"+file)
        doc.comment = check_sum
        doc_list.append(doc)
    return doc_list



from spdxmerge.SPDX_ShallowMerge import SPDX_ShallowMerger

from spdx.creationinfo import Person
from spdx.document import Document
from spdx.license import License
from spdx.utils import NoAssert
from spdx.version import Version


class Test_SPDX_ShallowMerger:
    def setup_method(self):
        doc1 = Document(
                name="Test document 1",
                spdx_id="SPDXRef-DOCUMENT1",
                namespace="http://example.com/spdx",
                data_license=License.from_identifier("CC0-1.0")
            )
        doc1.creation_info.add_creator(Person("John Smith", "john@example.com"))
        doc1.comment = "comment1"
        doc2 =  Document(
                name="Test document 2",
                spdx_id="SPDXRef-DOCUMENT2",
                namespace="http://example.com/spdx",
                data_license=License.from_identifier("CC0-1.0")
            )
        doc2.creation_info.add_creator(Person("John Smith", "john@example.com"))
        doc2.comment = "comment2"
        self.docs = [doc1,doc2]
        self.m = SPDX_ShallowMerger(self.docs, "http://example.com/spdx", "Test document", "John Doe")

    def test_document_creationInfo(self):
        self.m.doc_creationInfo()
        doc = self.m.get_document()

        assert doc.name == "Test document"
        assert doc.version == Version(2,3)
        assert doc.spdx_id == "http://example.com/spdx#SPDXRef-DOCUMENT"
        assert doc.namespace == "http://example.com/spdx"
        assert doc.data_license.identifier == "CC0-1.0"
        assert len(doc.creation_info.creators) == 1
        assert doc.creation_info.creators[0].name == "John Doe"

    def test_document_externalReference(self):
        self.m.doc_creationInfo()
        self.m.doc_externalDocumentRef()
        doc = self.m.get_document()

        assert len(doc.packages) == 1
        package = doc.packages[0]
        assert package.name == "Test document"
        assert package.version == "1.0"
        assert package.spdx_id == "http://example.com/spdx#SPDXRef-DOCUMENT"
        assert isinstance(package.download_location, NoAssert)
        assert len(doc.ext_document_references) == 2
        assert doc.ext_document_references[0].external_document_id == "SPDXRef-DOCUMENT1"
        assert doc.ext_document_references[0].spdx_document_uri == "http://example.com/spdx"
        assert doc.ext_document_references[0].checksum.identifier.name == "SHA1"
        assert doc.ext_document_references[0].checksum.value == "comment1"
        assert doc.ext_document_references[1].external_document_id == "SPDXRef-DOCUMENT2"
        assert doc.ext_document_references[1].spdx_document_uri == "http://example.com/spdx"
        assert doc.ext_document_references[1].checksum.identifier.name == "SHA1"
        assert doc.ext_document_references[1].checksum.value == "comment2"



import hashlib

def sha256sum(filename):
    h  = hashlib.sha256()
    b  = bytearray(128*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()


def sha1sum(filename):
    h  = hashlib.sha1()
    b  = bytearray(128*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()



import codecs
import os
from spdx.writers.json import write_document as write_json_document, InvalidDocumentError as JsonInvalidDocumentError
from spdx.writers.tagvalue import write_document as write_tagvalue_document, InvalidDocumentError as TagvalueInvalidDocumentError
from spdx.parsers.loggers import ErrorMessages
from spdxmerge.SPDX_DeepMerge import SPDX_DeepMerger
from spdxmerge.SPDX_ShallowMerge import SPDX_ShallowMerger

def create_merged_spdx_document(doc_list, docnamespace, name, author, email, merge_type):
    if merge_type == "deep":
        merger = SPDX_DeepMerger(doc_list, docnamespace, name, author, email)
        merger.doc_creationinfo()
        merger.doc_packageinfo()
        merger.doc_fileinfo()
        merger.doc_snippetinfo()
        merger.doc_other_license_info()
        merger.doc_relationship_info()
    elif merge_type == "shallow":
        merger = SPDX_ShallowMerger(doc_list, docnamespace, name, author, email)
        merger.doc_creationInfo()
        merger.doc_externalDocumentRef()

    return merger.get_document()

def write_file(doc, filetype, merge_type, outpath = None):
    result_filetype = "spdx" if filetype.lower() == "t" else "json"
    file = f"merged-SBoM-{merge_type}.{result_filetype}"
    if outpath:
        file = os.path.join(outpath, file)
    with codecs.open(file, mode="w", encoding="utf-8") as out:
        try:
            if result_filetype == "spdx":
                write_tagvalue_document(doc, out)
            else:
                write_json_document(doc, out)
        except (TagvalueInvalidDocumentError, JsonInvalidDocumentError) as e:
            print("Document is Invalid:\n\t", end="")
            print("\n\t".join(e.args[0]))
            messages = ErrorMessages()
            doc.validate(messages)
            print("\n".join(messages.messages))
        print("File "+file+" is generated")



# pylint: disable=wrong-import-position
import os
import sys
import click

sys.path.insert(0, "/".join(os.path.dirname(__file__).split('/')[:-1]))

from spdxmerge.SPDXMergeLib import create_merged_spdx_document, write_file
from spdxmerge.utils import read_docs



@click.command()
@click.option("--docpath", prompt="Directory path", required=True, help="Directory path with SPDX files to be merged")
@click.option("--outpath", prompt="Output directory path", required=False, prompt_required=False, help="Output directory path where merged file should be saved")
@click.option("--name", prompt="Product Name", required=True, help="Name of product for which SBoM is created")
@click.option("--mergetype", prompt="Shallow Merge -0 or Deep Merge-1", help="Enter 0 for shallow merge , 1 for deep merge", type=click.Choice(['0','1']), default='1')
@click.option("--author", prompt="SBoM Author name", required=True, help="Author who is writing SBoM")
@click.option("--email", prompt="SBoM author email address", help="Email address of the author")
@click.option("--docnamespace", prompt="Document namespace", help="URL where document is stored or organization URL", default="https://spdx.organization.name")
@click.option("--filetype", prompt="SBoM output file type SPDX tag value format - T or JSON - J",
              help="Enter T for SPDX tag value format, J for JSON", type=click.Choice(['T', 't', 'J','j']), default='J')
def main(docpath, name, mergetype, author, email, docnamespace, filetype, outpath = None):
    """Tool provides option to merge SPDX SBoM files. Provides two options for merging,
    Shallow Merge: New SBoM is created only with external ref links to SBoM files to be merged
    Deep Merge: New SBoM file is created by appending package, relationship, license information
    """
    doc_list = read_docs(docpath)
    merge_type = "shallow" if mergetype == '0' else "deep"
    doc = create_merged_spdx_document(doc_list, docnamespace, name, author, email, merge_type)
    write_file(doc, filetype, merge_type, outpath)


if __name__ == "__main__":
    main()



from spdx.creationinfo import Person
from spdx.license import License
from spdx.document import Document
from spdx.relationship import Relationship,RelationshipType
from spdx.version import Version

master_doc = Document()

class SPDX_DeepMerger():

    def __init__(self,doc_list=None,docnamespace=None,name=None,author=None,email=None):
        self.doc_list = doc_list
        self.docnamespace = docnamespace
        self.name = name
        self.author = author
        self.emailaddr = email

    def get_document(self):
        return master_doc

    def doc_creationinfo(self):
        master_doc.name = self.name
        master_doc.version = Version(2,3) # TODO Need to check from where to take this. can not hardcode here
        master_doc.spdx_id = self.docnamespace + "#SPDXRef-DOCUMENT"
        master_doc.namespace = self.docnamespace
        master_doc.data_license = License.from_identifier("CC0-1.0") #TODO Can not hardcode it here need to check from where to take it.
        master_doc.creation_info.add_creator(Person(self.author,self.emailaddr))
        master_doc.creation_info.set_created_now()

    def doc_packageinfo(self):
        """
        Append packges from document list
        """
        for doc in self.doc_list:
            master_doc.packages.extend(doc.packages)

    def doc_fileinfo(self):
        for doc in self.doc_list:
            master_doc.files.extend(doc.files)  #TODO Need to check this its not returning list

    def doc_snippetinfo(self):
        for doc in self.doc_list:
            master_doc.snippet.extend(doc.snippet)

    def doc_other_license_info(self):
        for doc in self.doc_list:
            master_doc.extracted_licenses.extend(doc.extracted_licenses)

    def doc_relationship_info(self):
        for doc in self.doc_list:
            # Add 'DESCRIBES' relationship between master and child documents, then import all relationships in child docs
            relationship = Relationship(master_doc.spdx_id+" "+RelationshipType.DESCRIBES.name+" "+doc.spdx_id)
            master_doc.add_relationship(relationship)
            master_doc.relationships.extend(doc.relationships)

    def doc_annotation_info(self):
        for doc in self.doc_list:
            master_doc.annotations.extend(doc.annotations)

    def doc_review_info(self):
        for doc in self.doc_list:
            master_doc.reviews.extend(doc.reviews)



import os

from setuptools import find_packages, setup

requirements_path = os.path.join(os.path.dirname(__file__), 'requirements.txt')

with open(requirements_path, "r", encoding="utf8") as f:
    required = f.read().splitlines()

setup(
    name='spdxmerge',
    version='0.2.0',
    description="merges content of two/more spdx sboms",
    long_description_content_type="text/markdown",
    url="https://github.com/philips-software/SPDXMerge",
    packages=find_packages(include=['spdxmerge'], exclude=['test', '*.test', '*.test.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=required,
    python_requires='>=3.8',
    entry_points='''
        [console_scripts]
        spdxmerge=spdxmerge.SPDXMerge:main
    '''
)



from datetime import datetime

from spdxmerge.SPDX_DeepMerge import SPDX_DeepMerger

from spdx.annotation import Annotation
from spdx.creationinfo import Person
from spdx.document import Document
from spdx.file import File
from spdx.license import License
from spdx.package import Package
from spdx.relationship import Relationship
from spdx.review import Review
from spdx.snippet import Snippet
from spdx.version import Version

class TestSPDXDeepMerger:

    def test_document_creation(self):
        doc1 = Document(
                name="Test document 1",
                spdx_id="SPDXRef-DOCUMENT1",
                namespace="http://example.com/spdx",
                data_license=License.from_identifier("CC0-1.0")
            )
        doc1.creation_info.add_creator(Person("John Smith", "john@example.com"))

        merger = SPDX_DeepMerger(doc_list=[doc1], docnamespace="https://example.com", name="Test Document", author="Test Author", email="test@example.com")
        merger.doc_creationinfo()

        doc = merger.get_document()
        assert doc.name == "Test Document"
        assert doc.version == Version(2,3)
        assert doc.spdx_id == "https://example.com#SPDXRef-DOCUMENT"
        assert doc.namespace == "https://example.com"
        assert doc.data_license == License.from_identifier("CC0-1.0")
        assert len(doc.creation_info.creators) == 1
        assert doc.creation_info.creators[0].name == "Test Author"

    def test_package_info(self):
        doc1 = Document()
        package1 = Package(name = "Package 1")
        doc1.add_package(package1)

        doc2 = Document()
        package2 = Package(name = "Package 2")
        doc2.add_package(package2)

        merger = SPDX_DeepMerger(doc_list=[doc1, doc2])
        merger.doc_packageinfo()

        doc = merger.get_document()
        assert len(doc.packages) == 2
        assert doc.packages[0].name == "Package 1"
        assert doc.packages[1].name == "Package 2"

    def test_file_info(self):
        doc1 = Document()
        file1 = File(name="file1")
        doc1.add_file(file1)

        doc2 = Document()
        file2 = File(name="file2")
        doc2.add_file(file2)

        merger = SPDX_DeepMerger(doc_list=[doc1, doc2])
        merger.doc_fileinfo()

        doc = merger.get_document()
        assert len(doc.files) == 2
        assert doc.files[0].name == "file1"
        assert doc.files[1].name == "file2"

    def test_snippet_info(self):
        doc1 = Document()
        snippet1 = Snippet()
        snippet1.comment = "snip1"
        doc1.add_snippet(snippet1)

        doc2 = Document()
        snippet2 = Snippet()
        snippet2.comment = "snip2"
        doc2.add_snippet(snippet2)

        merger = SPDX_DeepMerger(doc_list=[doc1, doc2])
        merger.doc_snippetinfo()

        doc = merger.get_document()
        assert len(doc.snippet) == 2
        assert doc.snippet[0].comment == "snip1"
        assert doc.snippet[1].comment == "snip2"

    def test_other_license_info(self):
        doc1 = Document()
        license1 = License.from_identifier("License1")
        doc1.add_extr_lic(license1)

        doc2 = Document()
        license2 = License.from_identifier("License2")
        doc1.add_extr_lic(license2)

        merger = SPDX_DeepMerger(doc_list=[doc1, doc2])
        merger.doc_other_license_info()

        doc = merger.get_document()
        assert len(doc.extracted_licenses) == 2
        assert doc.extracted_licenses[0].full_name == "License1"
        assert doc.extracted_licenses[0].identifier == "License1"
        assert doc.extracted_licenses[1].full_name == "License2"
        assert doc.extracted_licenses[1].identifier == "License2"

    def test_relationship_info(self):
        # Create two sample SPDX documents
        doc1 = Document()
        doc2 = Document()
        doc1.spdx_id = "SPDXRef-DOCUMENT1"
        doc2.spdx_id = "SPDXRef-DOCUMENT2"

        # Add relationship information to each document
        rel1 = Relationship("SPDXRef-PACKAGE1 DESCRIBES SPDXRef-DOCUMENT1")
        rel2 = Relationship("SPDXRef-PACKAGE2 DEPENDS_ON SPDXRef-PACKAGE1")
        doc1.add_relationship(rel1)
        doc2.add_relationship(rel2)

        # Merge the two documents using SPDX_DeepMerger
        merger = SPDX_DeepMerger([doc1, doc2], "https://example.com/spdx", "Test Document", "John Doe", "john@example.com")
        merger.doc_creationinfo()
        merger.doc_relationship_info()
        merged_doc = merger.get_document()

        # Verify that the merged document contains all the relationship information from both documents
        assert len(merged_doc.relationships) == 4
        assert merged_doc.relationships[0].relationship_type == "DESCRIBES"
        assert merged_doc.relationships[0].related_spdx_element == doc1.spdx_id
        assert merged_doc.relationships[0].spdx_element_id == "https://example.com/spdx#SPDXRef-DOCUMENT"

        assert merged_doc.relationships[1].relationship_type == "DESCRIBES"
        assert merged_doc.relationships[1].related_spdx_element == doc1.spdx_id
        assert merged_doc.relationships[1].spdx_element_id == "SPDXRef-PACKAGE1"

        assert merged_doc.relationships[2].relationship_type == "DESCRIBES"
        assert merged_doc.relationships[2].related_spdx_element == doc2.spdx_id
        assert merged_doc.relationships[2].spdx_element_id == "https://example.com/spdx#SPDXRef-DOCUMENT"

        assert merged_doc.relationships[3].relationship_type == "DEPENDS_ON"
        assert merged_doc.relationships[3].related_spdx_element == "SPDXRef-PACKAGE1"
        assert merged_doc.relationships[3].spdx_element_id == "SPDXRef-PACKAGE2"



    def test_annotation_info(self):
        # Create two sample SPDX documents
        doc1 = Document()
        doc2 = Document()
        doc1.spdx_id = "SPDXRef-DOCUMENT1"
        doc2.spdx_id = "SPDXRef-DOCUMENT2"

        # Add annotation information to each document
        ann1 = Annotation(spdx_id="SPDXRef-PACKAGE1", comment="This is a comment", annotator="John Doe")
        ann2 = Annotation(spdx_id="SPDXRef-PACKAGE2", comment="Another comment", annotator="Jane Smith")
        doc1.annotations.append(ann1)
        doc2.annotations.append(ann2)

        # Merge the two documents using SPDX_DeepMerger
        merger = SPDX_DeepMerger([doc1, doc2], "https://example.com/spdx", "Test Document", "John Doe", "john@example.com")
        merger.doc_creationinfo()
        merger.doc_annotation_info()
        merged_doc = merger.get_document()

        # Verify that the merged document contains all the annotation information from both documents
        assert len(merged_doc.annotations) == 2
        assert merged_doc.annotations[0].comment == "This is a comment"
        assert merged_doc.annotations[0].annotator == "John Doe"
        assert merged_doc.annotations[1].comment == "Another comment"
        assert merged_doc.annotations[1].annotator == "Jane Smith"


    def test_review_info(self):
        # Create two sample SPDX documents
        doc1 = Document()
        doc2 = Document()
        doc1.spdx_id = "SPDXRef-DOCUMENT1"
        doc2.spdx_id = "SPDXRef-DOCUMENT2"

        # Add review information to each document
        rev1 = Review("Reviewer 1", datetime.date, "Looks good to me")
        rev2 = Review("Reviewer 2", datetime.date, "I have some concerns about this file")
        doc1.reviews.append(rev1)
        doc2.reviews.append(rev2)

        # Merge the two documents using SPDX_DeepMerger
        merger = SPDX_DeepMerger([doc1, doc2], "https://example.com/spdx", "Test Document", "John Doe", "john@example.com")
        merger.doc_creationinfo()
        merger.doc_review_info()
        merged_doc = merger.get_document()

        # Verify that the merged document contains all the review information from both documents
        assert len(merged_doc.reviews) == 2
        assert merged_doc.reviews[0].reviewer == "Reviewer 1"
        assert merged_doc.reviews[0].comment == "Looks good to me"
        assert merged_doc.reviews[1].reviewer == "Reviewer 2"
        assert merged_doc.reviews[1].comment == "I have some concerns about this file"
